{PURPOSE

This programme takes as its input James Breen's Kanjidic, and outputs a copy of
this with the fields stripped. The fields are ordered, and may be alligned into
columns by padding them with spaces. Further it can attempt to output a romaji
transliteration of the hiragana and katakana readings.

LICENCE
KDCol is issued under the terms of the GNU licence.

AUTHOR
KDCol has been written by Leo V. Tilson. lvtilson@hotmail.com

LANGUAGE
This programme has been written in Pascal and compiled using Borland Turbo Pascal 3.02

ISSUES
Kanjidic is issued in UNIX text format. This results in lines being terminated by 
Carriage Returns (CR) or Line Feeds (LF), I can never remember which, while KDCol
running under DOS expects lines to terminate with the CR LF pair. Kanjidics 
end-of-line should be changed before it can be used under DOS. Several utilities
exist to do this, but the method I used was simply to open kanjidic using Windows
Wordpad, and 'save as' in a txt format.


NUTS AND BOLTS

CONSTANTS
    LINELEN    =1000;                Max length of input line
    MAXTIMES   =6;                   Max no of each field per line
    MINCOL     =10;                  Min width of output columns, for readings
    MAXCOL     =30;                  Max width of output columns, for readings
    DEFCOLWIDTH=12;                  Standard width of columns

TYPES
    RomStrType  =string[3];
    LineType    =array[1..LINELEN] of char;
    RomaType    =array[164..165,161..243,1..2] of RomStrType;
    BufType     =array[1..MAXCOL] of integer;

VARIABLES
Text    infile,outfile   Input and output files   

        For the following variables n_x is the number of fields of type x to be output.
                                    w_x is the max field width to be output.

Integer n_A                  Breen Kanjidic line number.
        n_B                  Bushu number.
        n_C                  Classical radical number.
        n_DR                 Fr Joseph De Roo codes
        n_DK                 Jack Halpern Kanji Learners Dictionary
        n_DO                 P.G. O'Neill's Essential Kanji.
        n_E                  Kenneth G.Henshell Guide to Japanese.
        n_eng , w_eng        English readings
        n_F                  Rank frequency
        n_G                  Jouyou grade
        n_H                  Jack Halperns Modern Dictionary
        n_hir , w_hir        Hiragana kun readings
        n_IN                 Spahn & Hadamitsky Kanji & Kana
        n_I                  Spahn & Hadamitsky Kanji Dictionary.
        n_jis                JIS code
        n_K                  Gakken Kanji Dictionary.
        n_kan                Kanji character
        n_kat , w_kat        KATAKANA ON readings
        n_kun , w_kun        Romaji kun readings
        n_L                  James Heisig: Rembering the Kanji.
        n_M                  Morohashi Daikanwajiten
        n_N                  Andrew Nelson's Modern Dictionary.
        n_O                  P.G.O'Neill Japanese Names.
        n_ON  , w_ON         ROMAJI ON readings
        n_P                  SKIP pattern code
        n_Q                  Four corner code
        n_S                  Stroke count
        n_U                  Unicode
        n_V                  Andrew Nelson's New Dictionary.
        n_W   , w_W          Korean readings
        n_X                  Cross reference code.
        n_Y   , w_Y          Pinyin readings
        n_Z                  Mis-Classification code.

Integer pad                  Whether the fields should be padded.                
Char    sep                  Field seperator character. 
        line                 Array of characters to store line read from input file.            
        romaji               Kana to romaji conversion table.     
           romaji [164,n,1]  Standard  hiragana
           romaji [164,n,2]  Shortened hiragana
           romaji [165,n,1]  Standard  KATAKANA
           romaji [165,n,2]  Shortened KATAKANA   

*** PROCEDURE openfiles;
    Ask the names of and opens the input and output files.

*** PROCEDURE setkana (VAR romaji:romatype);
    Initialises romaji which contains the roman representation of the kana

*** FUNCTION UPPER(low:RomStrType):RomStrType;
      Converts a 3 element string to its upper case equivalent

*** PROCEDURE AskFields(VAR n_A   , n_B   , n_C   , n_DR , n_DK  , n_DO  , n_E   ,
                            n_eng , w_eng , n_F   , n_G  , n_H   , n_hir , w_hir ,
                            n_IN  , n_I   , n_jis , n_K  , n_kan , n_kat , w_kat ,
                            n_kun , w_kun , n_L   , n_M  , n_N   , n_O   , n_ON  ,
                            w_ON  , n_P   , n_Q   , n_S  , n_U   , n_V   , n_W   ,
                            w_W   , n_X   , n_Y   , w_Y  , n_Z                    :integer;
                        VAR pad                                                   :integer;
                        VAR sep                                                   :char    );
    Ask user which fields should be included in the output file. e.g. n_N is the number
    of Nelson indices to be output, which should be 0 or 1, and w_eng is the width of
    english columns which should be MINCOL to MAXCOL

*** FUNCTION getyn:integer;
    reads terminal until y or n is pressed. Returns 0 (No), or 1 (Yes)

*** FUNCTION getint(min,max:integer):Integer;
    reads terminal until an integer between min and max is entered. Returns integer.

*** FUNCTION getcol(number:integer):integer;
    Asks user the width of the columns. Returns an integer width.

*** PROCEDURE readline (VAR line:linetype; VAR len:integer);
    reads a line of the input file. Outputs it as 'line'. Contains 'len' characters.

The next 5 procedures are used to write to the output file. They take the following parameters:
      line         Array containing a line copied from the input file.
      len          The number of characters contained in line.
      ID,ID_1,ID_2 Character used to identify the field to be written.
      times        Number of fields of ID type to be written per line.
      width        Max number of characters to write per field.
      pad          Whether the output field should be padded with spaces.
      sep          Character to separate fields.
      romaji       Array containing kana to romaji conversion table
      n_xyz        Whether the field identified by ID is to be written.

*** PROCEDURE WriteKanji(VAR line:linetype; n_kan:integer);
    Writes the kanji character contained in the two bytes before the first space

*** PROCEDURE WriteJis(VAR line:linetype; n_jis:integer);
    Writes the JIS code contained in the four bytes after the first space

*** PROCEDURE Writekana(VAR line:linetype; len:integer; ID,times,width,pad:integer; sep:char);
    Searches for a kana string and writes it exactly as written in the input file.

*** PROCEDURE WriteRoma(VAR line:linetype; len:integer; ID,times,width,pad:integer; sep:char;
                       VAR romaji:romatype);
    Searches for a kana string and writes its romaji representation.

*** PROCEDURE WriteIndex(VAR line:linetype; len:integer; ID_1,ID_2:char; term:char; times,width,pad:integer; sep:char);
    scans line from left to right looking for the field identified by its having
    ID_1 and ID_2 as its first two characters. ? as ID_2 stands for don't care,
    ! an integer. The field should be terminated by the term character. Does this 'times' times.
    Writes max of width characters per field. line is declared as VAR to prevent needless replication.


*** FUNCTION translit (VAR buffer:buftype; ID,width:integer; VAR romaji:romatype):integer;
    Writes the kana stored in buffer as a romaji string. Returns the number of characters
    written to the output file.}

{*********************************************************************************************}

PROGRAM kdcol (input,output,infile,outfile);

CONST
    LINELEN    =1000;                {Max length of input line}
    MAXTIMES   =6;                   {Max no of each field per line}
    MINCOL     =10;                  {Min width of output columns, for readings}
    MAXCOL     =30;                  {Max width of output columns, for readings}
    DEFCOLWIDTH=12;                  {Standard width of columns , for readings}

TYPE
    RomStrType  =string[3];
    LineType    =array[1..LINELEN] of char;
    RomaType    =array[164..165,161..243,1..2] of RomStrType;
    BufType     =array[1..MAXCOL] of integer;

VAR
    infile,outfile                                              :text;

    n_A   , n_B   , n_C   , n_DR , n_DK  , n_DO  , n_E   ,
    n_eng , w_eng , n_F   , n_G  , n_H   , n_hir , w_hir ,
    n_IN  , n_I   , n_jis , n_K  , n_kan , n_kat , w_kat ,
    n_kun , w_kun , n_L   , n_M  , n_N   , n_O   , n_ON  ,
    w_ON  , n_P   , n_Q   , n_S  , n_U   , n_V   , n_W   ,
    w_W   , n_X   , n_Y   , w_Y  , n_Z                          :integer;

    pad                                                         :integer;
    sep                                                         :char;
    line                                                        :linetype;
    i,linenum,len                                               :integer;
    romaji                                                      :romatype;

{***}
   PROCEDURE openfiles;

   {Ask the names of and opens the input and output files.}

   VAR
   infilename,outfilename:packed array[1..60] of char;

   BEGIN {openfiles}

      REPEAT
         write('What is the name of the input file?  ');
         readln(infilename);
         {$i-}
         assign(infile,infilename);
         reset(infile);
         {$i+}
      UNTIL (IOresult=0);

      REPEAT
         write('What is the name of the output file? ');
         readln(outfilename);
         {$i-}
         assign(outfile,outfilename);
         rewrite(outfile);
         {$i+}
      UNTIL (IOresult=0);

   END; {openfiles}
{###}

{***}
   PROCEDURE setkana (VAR romaji:romatype);

   {Initialises romaji which contains the roman representation of the kana}

   VAR i    :integer;
       tmp  :RomStrType;

{***} FUNCTION UPPER(low:RomStrType):RomStrType;

      {Converts a 3 element string to its upper case equivalent}

      VAR i:integer;
          tmp:RomStrType;

      BEGIN {UPPER}
         tmp:=low;
         FOR i:= 1 to length(tmp) DO
            tmp[i]:=chr(ord(tmp[i])-32);
         UPPER:=tmp;
      END; {UPPER}
{###}

   BEGIN {setkana}

      {Initialise the romaji table. In EUC hiragana are represented by two bytes,
       the first being 164, the second being an integer from 161 to 243.
       Start with the Hiragana, stored in romaji[164,161,1] TO romaji[164,243,1]  }

                                                                                       {a  i  u  e  o}
      romaji [164,161,1]:='a';   romaji [164,162,1]:='a';
      romaji [164,163,1]:='i';   romaji [164,164,1]:='i';
      romaji [164,165,1]:='u';   romaji [164,166,1]:='u';
      romaji [164,167,1]:='e';   romaji [164,168,1]:='e';
      romaji [164,169,1]:='o';   romaji [164,170,1]:='o';
                                                                                       {ka ki ku ke ko}
                                                                                       {ga gi gu ge go}
      romaji [164,171,1]:='ka';  romaji [164,172,1]:='ga';
      romaji [164,173,1]:='ki';  romaji [164,174,1]:='gi';
      romaji [164,175,1]:='ku';  romaji [164,176,1]:='gu';
      romaji [164,177,1]:='ke';  romaji [164,178,1]:='ge';
      romaji [164,179,1]:='ko';  romaji [164,180,1]:='go';
                                                                                       {sa shi su se so}
                                                                                       {za ji  zu ze zo}
      romaji [164,181,1]:='sa';  romaji [164,182,1]:='za';
      romaji [164,183,1]:='shi'; romaji [164,184,1]:='ji';
      romaji [164,185,1]:='su';  romaji [164,186,1]:='zu';
      romaji [164,187,1]:='se';  romaji [164,188,1]:='ze';
      romaji [164,189,1]:='so';  romaji [164,190,1]:='zo';
                                                                                       {ta chi tsu te to}
                                                                                       {da ji  tsu de do}
      romaji [164,191,1]:='ta';  romaji [164,192,1]:='da';
      romaji [164,193,1]:='chi'; romaji [164,194,1]:='ji';
      romaji [164,195,1]:='tsu'; romaji [164,196,1]:='tsu';  romaji [164,197,1]:='zu';
      romaji [164,198,1]:='te';  romaji [164,199,1]:='de';
      romaji [164,200,1]:='to';  romaji [164,201,1]:='do';
                                                                                       {na ni nu ne no}
      romaji [164,202,1]:='na';
      romaji [164,203,1]:='ni';
      romaji [164,204,1]:='nu';
      romaji [164,205,1]:='ne';
      romaji [164,206,1]:='no';
                                                                                       {ha hi fu he ho}
                                                                                       {ba bi bu be bo}
                                                                                       {pa pi pu pe po}
      romaji [164,207,1]:='ha';  romaji [164,208,1]:='ba';   romaji [164,209,1]:='pa';
      romaji [164,210,1]:='hi';  romaji [164,211,1]:='bi';   romaji [164,212,1]:='pi';
      romaji [164,213,1]:='fu';  romaji [164,214,1]:='bu';   romaji [164,215,1]:='pu';
      romaji [164,216,1]:='he';  romaji [164,217,1]:='be';   romaji [164,218,1]:='pe';
      romaji [164,219,1]:='ho';  romaji [164,220,1]:='bo';   romaji [164,221,1]:='po';
                                                                                       {ma mi mu me mo}
      romaji [164,222,1]:='ma';
      romaji [164,223,1]:='mi';
      romaji [164,224,1]:='mu';
      romaji [164,225,1]:='me';
      romaji [164,226,1]:='mo';
                                                                                       {ya    yu    yo} {small}
                                                                                       {ya    yu    yo} {large}
      romaji [164,227,1]:='a';   romaji [164,228,1]:='ya';
      romaji [164,229,1]:='u';   romaji [164,230,1]:='yu';
      romaji [164,231,1]:='o';   romaji [164,232,1]:='yo';
             {27,229,231 are the small versions of ya yu yo. They are entered here as 
               a,u,o since they will always be preceded by one of the shortened kana.}
                                                                                       {ra ri ru re ro}
      romaji [164,233,1]:='ra';
      romaji [164,234,1]:='ri';
      romaji [164,235,1]:='ru';
      romaji [164,236,1]:='re';
      romaji [164,237,1]:='ro';
                                                                                        {wa          o }
      romaji [164,238,1]:='wa';  romaji [164,239,1]:='wa';
      romaji [164,240,1]:='x';   romaji [164,241,1]:='x';
      romaji [164,242,1]:='o';
                                                                                        {n}
      romaji [164,243,1]:='n';

      {So far we have stored the normal hiragana. Now have to enter shortened forms of them.
       These forms are those used when a hiragana is followed by a small ya yo or yu. Store
       these in romaji[164,161,2] TO romaji[164,243,2]. First replicate romaji [164,n,1] }

      for i:=161 TO 243 DO romaji [164,i,2]:=romaji [164,i,1];

      {And then enter the exceptions.}

      romaji [164,173,2]:='ky';  romaji [164,174,2]:='gy';
      romaji [164,183,2]:='sh';  romaji [164,184,2]:='j';
      romaji [164,193,2]:='ch';  romaji [164,194,2]:='ji';
      romaji [164,203,2]:='ny';
      romaji [164,210,2]:='hy';  romaji [164,211,2]:='by';   romaji [164,212,2]:='py';
      romaji [164,223,2]:='my';
      romaji [164,234,2]:='ry';

      {The katakana are identical to the hiragana but represented in UPPER case.}

      FOR i:= 161 TO 243 DO romaji [165,i,1] := UPPER(romaji [164,i,1]);
      FOR i:= 161 TO 243 DO romaji [165,i,2] := UPPER(romaji [164,i,2]);

      {The table ends up as follows:
       romaji [164,n,1]           Standard  hiragana
       romaji [164,n,2]           Shortened hiragana
       romaji [165,n,1]           Standard  KATAKANA
       romaji [165,n,2]           Shortened KATAKANA }

   END; {setkana}
{###}

{***}
   PROCEDURE AskFields(VAR n_A   , n_B   , n_C   , n_DR , n_DK  , n_DO  , n_E   ,
                           n_eng , w_eng , n_F   , n_G  , n_H   , n_hir , w_hir ,
                           n_IN  , n_I   , n_jis , n_K  , n_kan , n_kat , w_kat ,
                           n_kun , w_kun , n_L   , n_M  , n_N   , n_O   , n_ON  ,
                           w_ON  , n_P   , n_Q   , n_S  , n_U   , n_V   , n_W   ,
                           w_W   , n_X   , n_Y   , w_Y  , n_Z                    :integer;
                       VAR pad                                                   :integer;
                       VAR sep                                                   :char    );
   {Ask user which fields should be included in the output file. e.g. n_N is the number
    of Nelson indices to be output, which should be 0 or 1, and w_eng is the width of
    english columns which should be MINCOL to MAXCOL}

   VAR nofields,acceptDefWh:integer;

{***} FUNCTION getyn:integer;

      {reads terminal until y or n is pressed. Returns 0 (No), or 1 (Yes)}

      VAR ch:char;

      BEGIN {getyn}
         getyn:=0;
         REPEAT write(' (y/n) '); readln(ch) UNTIL ch in ['0'..'9','y','n','Y','N'];
         IF ch in ['y','Y','1'..'9'] THEN getyn:=1;
      END; {getyn}
{###}

{***} FUNCTION getint(min,max:integer):Integer;

      {reads terminal until an integer between min and max is entered}

      VAR int:integer;

      BEGIN {getint}
         REPEAT
            write('(',min,'..',max,') ');
            {$i-}
            readln(int);
            {$i+}
         UNTIL (ioresult=0) AND (int>=min) AND (int<=max);
         getint:=int;
      END; {getint}
{###}

{***} FUNCTION getcol(number:integer):integer;

      {Asks user the width of the columns. Returns an integer width.}

      BEGIN {getcol}
         getcol:=0;
         IF (number >0)
         THEN
            BEGIN
               write('      Width of Column? (',MINCOL,'..',MAXCOL,')     ');
               getcol:=getint(MINCOL,MAXCOL);
            END;
      END; {getcol}
{###}

   BEGIN {askfields}

      {First, Initiliase all the parameters}

      n_A   :=0; n_B   :=0; n_C   :=0; n_DR :=0; n_DK  :=0; n_DO  :=0; n_E   :=0;
      n_eng :=0; w_eng :=0; n_F   :=0; n_G  :=0; n_H   :=0; n_hir :=0; w_hir :=0;
      n_IN  :=0; n_I   :=0; n_jis :=0; n_K  :=0; n_kan :=0; n_kat :=0; w_kat :=0;
      n_kun :=0; w_kun :=0; n_L   :=0; n_M  :=0; n_N   :=0; n_O   :=0; n_ON  :=0;
      w_ON  :=0; n_P   :=0; n_Q   :=0; n_S  :=0; n_U   :=0; n_V   :=0; n_W   :=0;
      w_W   :=0; n_X   :=0; n_Y   :=0; w_Y  :=0; n_Z   :=0;

      REPEAT
         writeln;
         write('Include Kanji character?               ');        n_kan:=getyn;

         writeln;
         write('Include Readings?                      ');
         IF (getyn>0)
         THEN
            BEGIN
               write('   Accept Default widths of columns?   ');  acceptDefwh:=getyn;

               write('   Number of Katakana On readings?     ');  n_kat:=getint(0,MAXTIMES);
                                                             IF (acceptDefWh=1)
                                                             THEN w_kat:=DEFCOLWIDTH
                                                             ELSE w_kat:=getcol(n_kat);

               write('   Number of Hiragana kun readings?    ');  n_hir:=getint(0,MAXTIMES);
                                                             IF (acceptDefWh=1)
                                                             THEN w_hir:=DEFCOLWIDTH
                                                             ELSE w_hir:=getcol(n_hir);

               write('   Number of Romaji ON readings?       ');  n_ON :=getint(0,MAXTIMES);
                                                             IF (acceptDefWh=1)
                                                             THEN w_ON :=DEFCOLWIDTH
                                                             ELSE w_ON :=getcol(n_ON);

               write('   Number of Romaji kun readings?      ');  n_kun:=getint(0,MAXTIMES);
                                                             IF (acceptDefWh=1)
                                                             THEN w_kun:=DEFCOLWIDTH
                                                             ELSE w_kun:=getcol(n_kun);

               write('   Number of English readings?         ');  n_eng:=getint(0,MAXTIMES);
                                                             IF (acceptDefWh=1)
                                                             THEN w_eng:=DEFCOLWIDTH
                                                             ELSE w_eng:=getcol(n_eng);

               write('   Number of Korean readings?          ');  n_W  :=getint(0,MAXTIMES);
                                                             IF (acceptDefWh=1)
                                                             THEN w_W  :=DEFCOLWIDTH
                                                             ELSE w_W  :=getcol(n_W);

               write('   Number of Pinyin readings?          ');  n_Y  :=getint(0,MAXTIMES);
                                                             IF (acceptDefWh=1)
                                                             THEN w_Y  :=DEFCOLWIDTH
                                                             ELSE w_Y  :=getcol(n_Y);
            END;

         writeln;
         write('Include Descriptors?                               ');
         IF (getyn>0)
         THEN
            BEGIN
               write('   Include Bushu number?               ');  n_B  :=getyn;
               write('   Include Classical Radical Number?   ');  n_C  :=getyn;
               write('   Include SKIP Codes?                 ');  n_P  :=getyn;
               write('   Include Four Corner Code?           ');  n_Q  :=getyn;
               write('   Number of stroke counts?            ');  n_S  :=getint(0,MAXTIMES);
               write('   Include De Roo Codes?               ');  n_DR :=getyn;
            END;

         writeln;
         write('Include Computer Codes?                ');
         IF (getyn>0)
         THEN
            BEGIN
               write('   Include jis codes?                  ');  n_jis:=getyn;
               write('   Include Unicode?                    ');  n_U  :=getyn;
            END;

         writeln;
         write('Include Dictionary Indices?            ');
         IF (getyn>0)
         THEN
            BEGIN
               write('   Include Breen Kanjidic Line No      ');  n_A  :=getyn;
               write('   Include DaiKan WaJiten Index?       ');  n_M  :=getyn;
               write('   Include Gakken Kanji Index?         ');  n_K  :=getyn;
               write('   Include Halpern Modern Index?       ');  n_H  :=getyn;
               write('   Include Halpern Learners Index?     ');  n_DK :=getyn;
               write('   Include Heisig Index?               ');  n_L  :=getyn;
               write('   Include Henshall Index?             ');  n_E  :=getyn;
               write('   Include Nelson Modern index?        ');  n_N  :=getyn;
               write('   Include Nelson New Index?           ');  n_V  :=getyn;
               write('   Include S & H Kanji & Kana Index?   ');  n_IN :=getyn;
               write('   Include S & H Kanji Dictionary ?    ');  n_I  :=getyn;
               write('   Include O''Neills Japanese Names?    '); n_O  :=getyn;
               write('   Include O''Neills Essential Names?   '); n_DO :=getyn;

            END;

         writeln;
         write('Include JoYo grade?                    ');        n_G  :=getyn;
         write('Include Frequency Rank?                ');        n_F  :=getyn;
         write('Number of Cross Reference Codes?       ');        n_X  :=getint(0,MAXTIMES);
         write('Number of Error Codes?                 ');        n_Z  :=getint(0,MAXTIMES);

         nofields:=n_A   + n_B  + n_C  + n_DR  + n_DK + n_DO  + n_E   + n_eng + n_F + n_G + n_H +
                   n_hir + n_IN + n_I  + n_jis + n_K  + n_kan + n_kat + n_kun + n_L + n_M + n_N +
                   n_O   + n_ON + n_P  + n_Q   + n_S  + n_U   + n_V   + n_W   + n_X + n_Y + n_Z;

         writeln;

         IF (nofields=0)
         THEN writeln(' Nothing to do !')
         ELSE writeln(nofields:3,' Fields to be written');

      UNTIL (nofields>0);

      writeln;
      write('Should the columns be padded?          ');        pad  :=getyn;
      write('Character to separate the columns?            '); readln(sep);

   END; {askfields}
{###}

{***}
   PROCEDURE readline (VAR line:linetype; VAR len:integer);

   {reads a line of the input file}

   VAR i:integer;

   BEGIN {readline}
      FOR i:=1 TO LINELEN DO line[i]:=' ';
      len:=1;
      REPEAT
         read(infile,line[len]);
         len:=len+1;
      UNTIL (EOLN(infile) OR (len=LINELEN));
   END; {readline}
{###}

{***}
   PROCEDURE WriteKanji(VAR line:linetype; n_kan:integer);

   {Writes the kanji character contained in the two bytes before the first space}

   VAR i:integer;

   BEGIN {writekanji}
      IF (n_kan=1)
      THEN
         BEGIN
            {Find the first space in the line...}
            i:=1; REPEAT i:=i+1 UNTIL (line[i]=' ');
            {...And write the two bytes before it}
            write(outfile,line[i-2],line[i-1],sep);
         END;
   END; {writekanji}
{###}

{***}
   PROCEDURE WriteJis(VAR line:linetype; n_jis:integer);

   {Writes the JIS code contained in the four bytes after the first space}

   VAR i,j:integer;

   BEGIN {writejis}
      IF (n_jis>0)
      THEN
         BEGIN
            {Search for the first space in the line...}
            i:=1; REPEAT i:=i+1 UNTIL (line[i]=' ');
            write(outfile,'JIS',sep);
            {...And write the four bytes after it}
            FOR j:=i+1 TO i+4 DO write(outfile,line[j]);
            write(outfile,' ',sep);
         END;
   END; {writejis}
{###}

{***}
   PROCEDURE Writekana(VAR line:linetype; len:integer; ID,times,width,pad:integer; sep:char);

   {Searches for a kana string and writes it exactly as written in the input file.}

   VAR
      charid:char;
      i,CountTimes,iout,k : integer;
      found               : boolean;

   BEGIN {writekana}

      i:=4;                                                                  {Start search at line[4]}
      charid:=chr(ID);
      FOR CountTimes:=1 TO times DO                                       {Search for 'times' strings}
         BEGIN                                                            {search for kana characters}
            found:=FALSE;
            REPEAT
               i:=i+1;
               found:= ( (line[i-1]=' ') AND (line[i]=charid) );
            UNTIL (found OR (i>=len) ) ;

            iout:=0;
            IF found
            THEN
               BEGIN
                                                            {First check for T1 or T2 before character}
                                                                           {and write out the T column}
                  k:=i-6;
                  REPEAT k:=k+1; UNTIL (k=i) OR (line[k] in ['T','t']);
                  IF (line[k] in ['T','t']) AND (line[k-1] = ' ') AND (line[k+1] in ['0'..'9'])
                  THEN write(outfile,line[k],line[k+1],sep)
                  ELSE write(outfile,' '    ,' '      ,sep);

                  REPEAT                                                    {Then write out the string}
                     write(outfile,line[i]);
                     i   :=i   +1;
                     iout:=iout+1;
                  UNTIL (line[i]=' ') OR (iout=width) OR (i>len);

               END
            ELSE
               write(outfile,' ',' ',sep);                                   {write the empty T column}

            IF (pad=1) THEN FOR k:= (iout+1) TO width DO write(outfile,' ');           {Pad out column}
            write(outfile,sep);
         END;

   END; { writekana}
{###}

{***}
   PROCEDURE WriteRoma(VAR line:linetype; len:integer; ID,times,width,pad:integer; sep:char;
                       VAR romaji:romatype);
   {Searches for a kana string and writes its romaji representation.}

   VAR
      charid:char;
      i,CountTimes,iout,k : integer;
      found               : boolean;
      buffer              : buftype;
      wsf                 : integer;

{***} FUNCTION translit (VAR buffer:buftype; ID,width:integer; VAR romaji:romatype):integer;
      {Writes the kana stored in buffer as a romaji string. Returns the number of characters
       written to the output file.}

      VAR
         i,wsf:integer;

      BEGIN {translit}

         wsf:=0;                                 {width so far. Nothing written to output file}
         i:=1;
         REPEAT
            IF (buffer[i]=ID)
            THEN
               BEGIN                                               {have found a kana character}
                  IF buffer[i+3] in [227,229,231]
                  THEN                               {kana character followed by small yo,ya,yu}
                     BEGIN
                        {write short form romaji}
                        write(outfile, romaji[ID,buffer[i+1],2]);
                        wsf:=wsf+length( romaji[ID,buffer[i+1],2]);
                     END
                  ELSE
                     BEGIN
                        {write standard romaji}
                        write(outfile, romaji[ID,buffer[i+1],1]);
                        wsf:=wsf+length( romaji[ID,buffer[i+1],1]);
                     END;
                  i:=i+2;
               END
            ELSE
               BEGIN                                              {have found a normal character}
                  write(outfile,chr(buffer[i]));
                  wsf:=wsf+1;
                  i:=i+1;
               END;

         UNTIL ( (buffer[i]=0) OR (wsf> (width-1)) );
         translit:=wsf;

      END; {translit}
{###}

   BEGIN  {writeroma}

      i:=4;                                                               {Start search at line[4]}
      charid:=chr(ID);
      FOR CountTimes:=1 TO times DO                                     {Search for times strings.}
         BEGIN                                                         {search for kana characters}
            found:=FALSE;
            REPEAT
               i:=i+1;
               found:= ( (line[i-1]=' ') AND (line[i]=charid) );
            UNTIL (found OR (i>=len) ) ;

            IF found
            THEN
               BEGIN                                                       {write out the characters}
                  iout:=1;
                                                      {but first check for T1 or T2 before character}
                                                                                   {and write it out}
                  k:=i-6;
                  REPEAT k:=k+1; UNTIL (k=i) OR (line[k] in ['T','t']);
                  IF (line[k] in ['T','t']) AND (line[k-1] = ' ') AND (line[k+1] in ['0'..'9'])
                  THEN write(outfile,line[k],line[k+1],sep)
                  ELSE write(outfile,' '    ,' '      ,sep);
                                                                    {copy the whole string to buffer}
                  FOR k:= 1 TO MAXCOL DO buffer[k]:=0;
                  REPEAT
                     buffer[iout]:=ord(line[i]);
                     i   :=i   +1;
                     iout:=iout+1;
                  UNTIL (line[i]=' ') OR (iout=MAXCOL) OR (i>len);
                  wsf:=translit(buffer,ID,width,romaji);                        {then write out the string}
               END
            ELSE
               BEGIN                                                    {write out the empty T column}
                  write(outfile,' ',' ',sep);
                  wsf:=0;
               END;
                                                                                  {pad out the column}
            IF (pad=1) THEN FOR k:= (wsf+1) TO width DO write(outfile,' ');
            write(outfile,sep);
         END;
   END; { writeroma}

{***}
   PROCEDURE WriteIndex(VAR line:linetype; len:integer; ID_1,ID_2:char; term:char; times,width,pad:integer; sep:char);

   {scans line from left to right looking for the field identified by its having
    ID_1 and ID_2 as its first two characters. ? as ID_2 stands for don't care,
    ! an integer. The field should be terminated by the term character. Does this 'times' times.
    Writes max of width characters per field. line is declared as VAR to prevent needless replication.}

   VAR
      i,CountTimes,iout,k : integer;
      found               : boolean;

   BEGIN {writeindex}
      i:=4;

      IF (ID_2='!')                             {If the character after ID_1 should be an integer}
      THEN
         BEGIN
            IF (times>0) THEN write(outfile,ID_1,sep);                         {Write field identifier}
                                                                             {search for times strings}
            FOR CountTimes:=1 TO times DO
               BEGIN
                                                                                    {search for string}
                  found:=FALSE;
                  REPEAT
                     i:=i+1;
                     found:= ( (line[i-1]=' ') AND (line[i]=ID_1) AND (line[i+1] in ['0'..'9']) );
                  UNTIL (found OR (i>=len) ) ;

                  iout:=0;
                  IF found                                                                {write it out}
                  THEN
                     REPEAT
                        i   :=i   +1;
                        iout:=iout+1;
                        write(outfile,line[i]);
                     UNTIL (line[i+1]=term) OR (iout=width) OR (i>len);

                  IF (pad=1) THEN FOR k:= (iout+1) TO width DO write(outfile,' ');           {pad column}
                  write(outfile,sep);
               END;
         END;

      IF (ID_2='?')                             {If we don't care what the character after ID_2 should be}
      THEN
         BEGIN
            IF (times>0) THEN write(outfile,ID_1,sep);                             {write field identifier}
                                                                                 {search for times strings}
            FOR CountTimes:=1 TO times DO
               BEGIN
                                                                                        {search for string}
                  found:=FALSE;
                  REPEAT
                     i:=i+1;
                     found:= ( (line[i-1]=' ') AND (line[i]=ID_1) );
                  UNTIL (found OR (i>=len) ) ;

                  iout:=0;
                  IF found                                                                  {write it out}
                  THEN
                     REPEAT
                        i   :=i   +1;
                        iout:=iout+1;
                        write(outfile,line[i]);
                     UNTIL (line[i+1]=term) OR (iout=width) OR (i>len);

                  IF (pad=1) THEN FOR k:= (iout+1) TO width DO write(outfile,' ');            {pad column}
                  write(outfile,sep);
               END;
         END;

      IF NOT(ID_2 in ['!','?'])                       {If we do care what the character after ID_2 should be}
      THEN
         BEGIN
            IF (times>0) THEN write(outfile,ID_1,ID_2,sep);                          {write field identifier}

            FOR CountTimes:=1 TO times DO
               BEGIN
                                                                                        {search for string}
                  found:=FALSE;
                  REPEAT
                     i:=i+1;
                     found:= ( (line[i-1]=' ') AND (line[i]=ID_1) AND (line[i+1]=ID_2) );
                  UNTIL (found OR (i>=len) ) ;
                  i:=i+1;
                  iout:=0;
                  IF found                                                                  {write it out}
                  THEN
                     REPEAT
                        i   :=i   +1;
                        iout:=iout+1;
                        write(outfile,line[i]);
                     UNTIL (line[i+1]=term) OR (iout=width) OR (i>len);

                  IF (pad=1) THEN FOR k:= (iout+1) TO width DO write(outfile,' ');            {pad column}
                  write(outfile,sep);
               END;
         END;

   END; {writeindex}
{###}

BEGIN {kanjicol}

   openfiles;

   AskFields(n_A   , n_B   , n_C   , n_DR , n_DK  , n_DO  , n_E   ,
             n_eng , w_eng , n_F   , n_G  , n_H   , n_hir , w_hir ,
             n_IN  , n_I   , n_jis , n_K  , n_kan , n_kat , w_kat ,
             n_kun , w_kun , n_L   , n_M  , n_N   , n_O   , n_ON  ,
             w_ON  , n_P   , n_Q   , n_S  , n_U   , n_V   , n_W   ,
             w_W   , n_X   , n_Y   , w_Y   , n_Z  , pad   , sep    );

   setkana(romaji);

                                           {Read and write the copyright notice}
   readline(line,len);
   FOR i:= 1 TO len DO write(outfile,line[i]);writeln(outfile);

                                        {Then loop through the rest of the lines}
   linenum:=0;

   WHILE NOT EOF(infile) DO
      BEGIN
         linenum:=linenum+1;
         if ((linenum mod 100)=0)THEN writeln('Writing line: ',linenum:5);

         readline (line,len);
                               {The following lines determine the order in which the
                       fields in the output file are arranged. They may be rearranged
                              if desired since they are not dependent on one another}

                                                             { Write Kanji character}
         WriteKanji(line,n_kan);
                                                                     {Write readings}
         WriteKana(line  , len ,164 , n_hir , w_hir , pad , sep);
         WriteKana(line  , len ,165 , n_kat , w_kat , pad , sep);
         WriteRoma(line  , len ,164 , n_kun , w_kun , pad , sep , romaji );
         WriteRoma(line  , len ,165 , n_ON  , w_ON  , pad , sep , romaji );
         WriteIndex(line , len , '{' , '?' , '}' , n_eng, w_eng , pad , sep);
         WriteIndex(line , len , 'Y' , '?' , ' ' , n_Y  , w_y   , pad , sep);
         WriteIndex(line , len , 'W' , '?' , ' ' , n_W  , w_W   , pad , sep);

                                                                   {Write descriptors}
         WriteIndex(line , len , 'B' , '?' , ' ' , n_B  ,  4    , pad , sep);
         WriteIndex(line , len , 'C' , '?' , ' ' , n_C  ,  4    , pad , sep);
         WriteIndex(line , len , 'P' , '?' , ' ' , n_P  , 10    , pad , sep);
         WriteIndex(line , len , 'Q' , '?' , ' ' , n_Q  , 10    , pad , sep);
         WriteIndex(line , len , 'S' , '?' , ' ' , n_S  ,  4    , pad , sep);
         WriteIndex(line , len , 'D' , 'R' , ' ' , n_DR , 10    , pad , sep);

                                                                 {Write computer codes}
         writeJis  (line , n_jis);
         WriteIndex(line , len , 'U' , '?' , ' ' , n_U  , 10     , pad , sep);

                                                              {Write dictionary indices}
         IF(n_A=1) THEN write(outfile , 'A' , sep , linenum:5   , sep);
         WriteIndex(line , len , 'D' , 'K' , ' ' , n_DK , 10    , pad , sep);
         WriteIndex(line , len , 'D' , 'O' , ' ' , n_DO , 10    , pad , sep);
         WriteIndex(line , len , 'E' , '?' , ' ' , n_E  , 10    , pad , sep);
         WriteIndex(line , len , 'H' , '?' , ' ' , n_H  , 10    , pad , sep);
         WriteIndex(line , len , 'I' , '!' , ' ' , n_I  , 10    , pad , sep);
         WriteIndex(line , len , 'I' , 'N' , ' ' , n_IN , 10    , pad , sep);
         WriteIndex(line , len , 'K' , '?' , ' ' , n_K  , 10    , pad , sep);
         WriteIndex(line , len , 'L' , '?' , ' ' , n_L  , 10    , pad , sep);
         WriteIndex(line , len , 'N' , '?' , ' ' , n_N  , 10    , pad , sep);
         WriteIndex(line , len , 'M' , 'N' , ' ' , n_M  , 10    , pad , sep);
         WriteIndex(line , len , 'M' , 'P' , ' ' , n_M  , 10    , pad , sep);
         WriteIndex(line , len , 'V' , '?' , ' ' , n_V  , 10    , pad , sep);

                                                             {write miscellaneous items}
         WriteIndex(line , len , 'F' , '?' , ' ' , n_F  , 10    , pad , sep);
         WriteIndex(line , len , 'G' , '?' , ' ' , n_G  , 10    , pad , sep);
         WriteIndex(line , len , 'X' , '?' , ' ' , n_X  , 10    , pad , sep);
         WriteIndex(line , len , 'Z' , '?' , ' ' , n_Z  , 10    , pad , sep);

         writeln(outfile);

      END; {while}


   writeln(linenum:6,' lines written');
   close(infile);
   close(outfile);

END. {kanjicol}

